# coding=utf-8
import sys
from typing import Any, Dict, List, Optional
from mammut.apps.corpus_compiler_service import CorpusCompilerService
from mammut.common.basic_tokenizer import BasicTokenizer
from mammut.common.corpus.corpus_map import CorpusMap
from mammut.common.indexer import CorpusIndexer, NGramIndexer
from mammut.common.lexicon.linguistic_standard import LinguisticStandard
from mammut.common.lexicon.dictionary import Dictionary
from mammut.common.lexicon.paradigm_producer import (
    ParadigmProducer,
    ParadigmProducerGroup,
)
from mammut.common.storage.storage_base import SheetsStorageBase
from mammut.linguistics.concordance import Concordance
from mammut.common.annotations import Restitution, MammutEventAnnotation
import mammut.visualization.lexicalization_widget as lexicalization_widget
import mammut.visualization.disambiguation_widget as disambiguation_widget
import mammut.visualization.annotations_widget as annotations_widget
import mammut.visualization.annotations_validator_widget as annotations_validator_widget
import mammut.visualization.widgets_utilities as widgets_utilities
from mammut.common.annotations import MammutAnnotations
from mammut.common.annotations_validator import CorpusAnnotationsValidator
from mammut.common.ngram import NGram
from mammut.common.corpus_transformer import *
from mammut.common.synthetic.synthetic_corpus_generator.morphemes_generator import (
    MorphemesGenerator,
)
from mammut.common.synthetic.synthetic_corpus_generator.synthetic_corpus_generator import (
    SyntheticCorpusGenerator,
)
from mammut.common.catalog.catalog_manager import Catalog
import logging
from IPython.display import display
from ipywidgets import widgets
import pandas as pd

from mammut.common.storage.storage_manager import StorageManager

log = logging.getLogger(__name__)


class Package:
    """The entrypoint class for Mammut Computational Linguistics Framework.
    Package class provides a client API for Computational Linguistics
    operations to work with structured, semi-structured and unstructured
    information supported by Mammut Framework.

    Args:
        package_id(str): ID of the presentation document. It can be a Google
            presentation ID or a local path to a .pptx file.
        standard_spreadsheet_id(str): ID of the standard document. It can be a Google
            presentation ID or a local path to a .pptx file.
        package_id_is_spreadsheet(bool): If True, package_id arg can be
            directly an ID of a package spreadsheet file.
    """

    storage_manager = StorageManager()

    def __init__(
        self,
        package_id: str,
        standard_spreadsheet_id: str,
        package_id_is_spreadsheet: bool = False,
        show_progress: bool = True,
        standard_base_sheet_title: str = "base",
        standard_dict_sources_sheet_title: str = "dict-sources",
        standard_pos_sheet_title: str = "pos-tags",
        standard_features_sheet_title: str = "features",
        standard_dimensions_sheet_title: str = "dimensions",
        dictionary_sheet_title: str = "dictionary",
        standard_functionals_sheet_title: str = "functionals",
        standard_non_functionals_sheet_title: str = "non-functionals",
        standard_primes_sheet_title: str = "primes",
        standard_synonym_sheet_title: str = "synonym",
        synonym_sheet_title: str = "synonym"
    ):
        # Start package loading
        log.info("Start loading package...")
        self.package_id_is_spreadsheet = package_id_is_spreadsheet
        self.spreadsheet_id = self._extract_spreadsheet_id_from_slide(
            package_id
        )
        self.main_id = package_id
        self.package_name = self.spreadsheet_id.lower()
        self.show_progress: bool = show_progress
        self.standard_spreadsheet_id = standard_spreadsheet_id
        self.standard_base_sheet_title = standard_base_sheet_title
        self.standard_dict_sources_sheet_title = standard_dict_sources_sheet_title
        self.standard_pos_sheet_title = standard_pos_sheet_title
        self.standard_features_sheet_title = standard_features_sheet_title
        self.standard_dimensions_sheet_title = standard_dimensions_sheet_title
        self.dictionary_sheet_title = dictionary_sheet_title
        self.standard_functionals_sheet_title = standard_functionals_sheet_title
        self.standard_non_functionals_sheet_title = standard_non_functionals_sheet_title
        self.standard_primes_sheet_title = standard_primes_sheet_title
        self.standard_synonym_sheet_title = standard_synonym_sheet_title
        self.synonym_sheet_title = synonym_sheet_title
        self.storage_manager.infer_spreadsheet_storage(package_id)
        self.tokenizer = BasicTokenizer()
        self.standard: Optional[LinguisticStandard] = None
        self.dictionary: Optional[Dictionary] = None
        self.synonym_sheet_title: str = "synonym"
        self.jde = lexicalization_widget
        self.disambiguation_widget = disambiguation_widget
        self.annotations_widget = annotations_widget
        self.annotations_validator_widget = annotations_validator_widget
        self.jde.DEFAULT_STANDARD = self.standard
        self.operation_counter = 0
        self.status_text_box = widgets.Text(
            value="Your package is being initialized!",
            disabled=True,
            layout=widgets.Layout(width="900px", height="120px"),
        )
        if self.show_progress:
            display(self.status_text_box)
        self.corpus_map: Optional[CorpusMap] = None
        self.corpus_indexer = CorpusIndexer()
        log.info("Indexer loaded")
        self.ngram_indexer = NGramIndexer()
        self.ngram_query = NGram()
        log.info("NGram handler loaded")
        self.concordance = Concordance()
        self.library = None
        self.restitution = Restitution(self, self.spreadsheet_id)
        log.info("Restitutions loaded")
        self.corpus_compiler_service = None
        self.paraphrases = None
        self.morphemes_generator: Optional[MorphemesGenerator] = None
        self.synthetic_corpus_generator: Optional[SyntheticCorpusGenerator] = None
        self.catalog: Optional[Catalog] = None
        self.annotations: Optional[MammutAnnotations] = None
        self.corpus_annotations_validator: Optional[CorpusAnnotationsValidator] = None
        log.info("Package loaded!!!")

    def load_main_classes(self):
        """Loads main classes LinguisticStandard, Dictionary, and CorpusMap.
        """
        self.standard = LinguisticStandard(
            self.standard_spreadsheet_id,
            self.standard_base_sheet_title,
            self.standard_dict_sources_sheet_title,
            self.standard_pos_sheet_title,
            self.standard_features_sheet_title,
            self.standard_dimensions_sheet_title,
            self.standard_functionals_sheet_title,
            self.standard_non_functionals_sheet_title,
            self.tokenizer,
            self.standard_primes_sheet_title,
            self.standard_synonym_sheet_title,
        )
        self.dictionary = Dictionary(
            self.spreadsheet_id,
            self.dictionary_sheet_title,
            self.synonym_sheet_title,
            self.tokenizer,
            self.standard,
            False,
            self.package_name,
        )
        self.corpus_map = CorpusMap(
            self.spreadsheet_id,
            self.tokenizer,
            self.dictionary,
            self.standard,
            self.catalog,
            self._update_status_text_box,
        )
        self._index_all_corpus()
        self.annotations = MammutAnnotations(self)
        self.corpus_annotations_validator = CorpusAnnotationsValidator(self)
        self.morphemes_generator = MorphemesGenerator(self.standard, self.dictionary)
        self.synthetic_corpus_generator = SyntheticCorpusGenerator(self.morphemes_generator)

    def _extract_spreadsheet_id_from_slide(self, presentation_id: str):
        """Extracts the spreadsheet_id from the presentation specified by
            presentation_id.

        Note:
            For local .pptx files, there may be runtime errors if the
            environment variable LOCAL_STORAGE_BASE_FOLDER is not set
            correctly. LOCAL_STORAGE_BASE_FOLDER refers to the local
            folder where the package resides.

        Args:
            presentation_id(str): Google presentation ID or local
                path to a .pptx file.

        Returns:
            If the presentation_id is a Google presentation, returns the
                spreadsheet_id from the Google Spreadsheet.

            If the presentation_id is a local path to a .pptx file, returns
                the absolute path to the .xlsx file found in the .pptx file
                entry for the spreadsheet_id.

            If self.package_id_is_spreadsheet, then the package_id class arg
                is returned immediately.
        """
        if self.package_id_is_spreadsheet:
            return presentation_id
        else:
            slide_elements = self.storage_manager.get_slide_elements(
                presentation_id
            )
            for pageElements in slide_elements:
                for element in pageElements:
                    if (
                        not re.search(
                            self.storage_manager.get_slides_metadata_re_string(
                                presentation_id
                            ),
                            element.content,
                        )
                        is None
                    ):
                        return re.search(
                            self.storage_manager.get_slides_metadata_re_string(
                                presentation_id
                            ),
                            element.content,
                        ).group(2)

    def get_custom_table(
        self,
        spreadsheet_id: str,
        sheet_title: str,
        start_cell: str = "A1",
        end_cell: str = "",
        transpose: bool = False,
    ) -> pd.DataFrame:
        """Get a Dataframe from a sheet in a spreadsheet.

        Todo:
            * Test method with local packages (.xlsx files)
            * Transpose support will be eventually removed.

        Args:
            spreadsheet_id(str): The ID of the Google Spreadsheet.
            sheet_title(str): sheet name in the Google Spreadsheet.
            start_cell(str): first cell of the range in A1 notation.
            end_cell(str): last cell of the range in A1 notation.
            transpose(bool): True if the sheet is written in transpose order.

        Returns:
            Returns the Pandas' Dataframe containing the specified sheet
                range.
        """
        data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
            spreadsheet_id,
            sheet_title,
            start_cell[0],
            start_cell[1 : len(start_cell)],
            end_cell,
            transpose,
        )
        return data_frame

    def set_custom_table(
        self,
        spreadsheet_id: str,
        data_frame,
        sheet_title: str,
        start_row=SheetsStorageBase.DEFAULT_START_ROW,
        start_col=SheetsStorageBase.DEFAULT_START_COLUMN,
        is_update_cells: bool = False,
    ) -> str:
        """Save a Dataframe to a sheet in a spreadsheet.

        Todo:
            * Test method with local packages (.xlsx files)

        Args:
            spreadsheet_id(str): The ID of the Google Spreadsheet.
            data_frame(Dataframe): The Pandas' Dataframe to save.
            sheet_title(str): sheet name in the Google Spreadsheet.
            start_row(str): start row number in the spreadsheet.
            start_col(str): start column name in the spreadsheet.
            is_update_cells(bool): False to append data_frame.
                True will override cells if needed.

        Returns:
            Returns a string containing the A1 notation of the modified
            spreadsheet.
        """
        rows = data_frame.to_numpy().tolist()
        if not is_update_cells:
            return self.storage_manager.add_rows_to_spreadsheet(
                spreadsheet_id, sheet_title, rows, start_row, start_col
            )
        else:
            return self.storage_manager.update_rows_from_spreadsheet(
                spreadsheet_id, sheet_title, rows, start_row, start_col
            )

    def _index_all_corpus(self):
        """Index all corpus.

        Index all corpus using the configured backend
        (Elasticsearch at the moment of the doc).
        This is used by several other components to perform fast lookups over
        the corpus.
        """
        for corpus in self.corpus_map.get_all_corpus():
            if corpus.indexable:
                self.corpus_indexer.index_corpus(
                    corpus.corpus_data_frame,
                    corpus.spreadsheet_id,
                    corpus.corpus_sheet_title,
                    self.package_name,
                )

    def check_concordance(self, query_list: List[str]) -> widgets.HTML:
        """Checks concordance in a corpus.
            
        This method returns the concordance of one or more words
        in context of the events from the corpus noted in the corpus_map
        of the spreadsheet. A corpus is a collection of prototypical 
        conversations, while events are each of the messages in those
        conversations.

        Example:
            >>> input=(['restaurant'])
            >>> check_concordance(input)
            [
             this ____ restaurant, apart from being emblematic is affordable',
            'the ____ restaurant is very nice, they have a nice atmosphere',
            'I had very good recommendations from this ___________ restaurant'
            ]
            
        Args:
            query_list (List[str]): list of lemmas to check concordance for.
                
        Returns:
            Returns all concordance occurrences of one or more words in
                HTML format.
        """
        concordance_results = []
        for query in query_list:
            concordance_results.append(
                self.concordance.get_data_from_index(self.package_name, query)
            )
        concordance_html = lexicalization_widget.concordance_template.render(
            {"concordance_results": concordance_results}
        )
        return widgets.HTML(value=concordance_html)

    def check_concordance_for_lemma(self, lemma: str) -> widgets.HTML:
        """Checks the concordance of a lemma in the corpus. 
        
        This method returns the concordance of a specific lemma
        in context of the events from the corpus (C, N, M) noted in the
        corpus_map of the spreadsheet. Through this method developers can
        check all concordance occurrences of a lemma, also, they can see all
        the inflections of a lemma that exist in events of the corpus
        declared. A corpus is a collection of prototypical conversations,
        while events are each of the messages in those conversations.

        Example:
            >>> input='have'
            >>> check_concordance_for_lemma(input)
            ['we do not _____ have this service',
            'do you _____ have a website ?',
            'in addition to our coffees we also _____ have natural juices']

        Args:
            lemma: word to check concordance for.

        Returns:
            Returns all concordance occurrences of a lexicalized lemmas in
                HTML format.
        """
        concordance_results = []
        query_list = self.dictionary.get_morphemes_of_lemma(lemma)
        for query in query_list:
            concordance_results.append(
                self.concordance.get_data_from_index(self.package_name, query)
            )
        concordance_html = lexicalization_widget.concordance_template.render(
            {"concordance_results": concordance_results}
        )
        return widgets.HTML(value=concordance_html)

    def show_defined_lemmas(self, date: str) -> pd.DataFrame:
        """This method shows lemmas or words defined in a specific date.

        Example:            
            >>> package.show_defined_lemmas('2019/05/20')

        Args:
            date (str): date to check. The date parameter must be a in the
                format ('YYYY/MM/DD') supported by Elasticsearch.
                https://www.elastic.co/guide/en/elasticsearch/reference/current/date.html

        Returns:
            Returns a Pandas' Dataframe with lemmas defined in the
            date specified.
        """
        l_d = self.dictionary.definition_indexer.get_defined_lemmas(date)
        return widgets_utilities.create_defined_words_section(l_d)

    def show_lexicalization_widget(self):
        """Shows the lexicalization widget view.
            
        The lexicalization widget is a tool that allows developers to create 
        an internal lexicon or dictionary from the words present in each of 
        the events of the corpus declared in the corpus map of a spreadsheet. 
        The words present in the events of a corpus are incorporated into 
        the internal dictionary through lexicalization.

        Returns:
            Displays a box containing various add-ons. 
            Through these add-ons it is possible to define a word.
            Returns the view of the lexicalization widget. 
        """
        all_box = lexicalization_widget.show_restart_add_lemma_entry_section_button(
            self.dictionary, self.package_name
        )

        def search_for_lemma(sender):
            """This method shows a box in the lexicalization widget that
            permits the developer to enter a specific word, and search a
            definition of it in the dictionaries available:
            Word-Reference, RAE, Oxford, or Wiktionary.
            """
            new_accordion = lexicalization_widget.create_external_dictionaries_section(
                all_box.children[1].children[0].children[0].value.lower(),
                self.dictionary.standard,
                all_box.children[1].children[0].children[3].value,
            )
            for i in range(0, len(all_box.children[1].children[1].children)):
                all_box.children[1].children[1].set_title(
                    i, new_accordion.get_title(i)
                )
            all_box.children[1].children[1].children = new_accordion.children

        def handle_restart_section_button_click(sender):
            """Handles the words browser in the lexicalization widget.
                
            Returns:
                Resets the section to search words in the lexicalization
                    widget.
            """
            all_box.children = [
                all_box.children[0],
                lexicalization_widget.create_lemma_section(
                    self.dictionary, self.package_name
                ),
            ]
            all_box.children[1].children[0].children[4].on_click(
                search_for_lemma
            )

        all_box.children[0].on_click(handle_restart_section_button_click)
        all_box.children[1].children[0].children[4].on_click(search_for_lemma)
        return all_box

    def _update_status_text_box(self, new_value: str = ""):
        """Updates the progress widget when the Package is loading.

        """
        self.operation_counter = self.operation_counter + 1
        standar_text = "Operation number: " + str(self.operation_counter)
        if new_value:
            value = standar_text + " --- extra info: " + new_value
        else:
            value = standar_text
        self.status_text_box.value = value

    def add_restitution(
        self, event_annotation: MammutEventAnnotation, restitution_input: str
    ):
        """Adds a word in a corpus event.
            
        The 'add restitution' component is a section that is part of 
        the annotation widget. Through this component, developers can add 
        a word which is also part of a specific event in a corpus, 
        but that word does not appear in the event explicitly.
        The word that does not appear in the event explicitly is an ellipsis.
        Ellipsis is the omission from a clause of one 
        or more words that are nevertheless understood 
        in the context of the remaining elements.
        This component allows developers to add the
        restitution in the initial position of the event. 
        This method is used directly by the annotation widget. 
        The annotation widget has the data that points to the corpus, 
        scenario, and corpus event that requires restitution. 
        The annotation widget passes the required MammutEventAnnotation 
        given user interaction.

        #TODO: Example must use MammutEventAnnotation class
        Example:
            >>> input='Just saying!'
            >>> resitution='I'
            >>> add_restitution(input, restitution)
            'I just saying'
        
        TODO: 
            * Currently, this method only places the restitution
            in the initial position of the event. 
            * The method should allow developers to choose
            the location of the event where restitution is needed. 

        Args:
            event_annotation (MammutEventAnnotation): corpus event to be
                annotated.
            restitution_input (str): the word that does not appear in the
                event explicitly. (Ellipsis)

        """
        deleted_restitutions = self.delete_restitutions(event_annotation)
        restitutions = [restitution_input.lower()] + deleted_restitutions
        ordered_restitutions = restitutions[::-1]
        for rest in ordered_restitutions:
            self.restitution.add_restitution(
                event_annotation.event["corpus"]["name"],
                event_annotation.event[
                    MammutEventAnnotation.SCENARIO_INDEX_ID
                ],
                event_annotation.event["event_entry"],
                event_annotation.event["tokenized_message_index"],
                rest,
            )
        event_annotation.update_documents_after_restitution(
            " ".join(restitutions)
        )
        self.annotations.delete_all_annotations(event_annotation)

    def delete_restitutions(
        self, event_annotation: MammutEventAnnotation
    ) -> List[str]:
        """Deletes all restored words in a corpus event.

        The 'delete restitution' component is a section that is part of 
        the annotation widget. Through this component, developers can delete 
        all words that were previously restored for a specific event in a
        corpus.

        #TODO: Example must use event annotation class
        Example:
            >>> input= 'I just saying!'
            >>> delete_restitution(input, restitution)
            'Just saying!'

        Args:
            event_annotation (MammutEventAnnotation): corpus event to be annotated.

        #TODO: Verify the meaning of the return value.
        Returns:
            Returns the number of deleted restitutions.
        """
        deleted_restitutions = self.restitution.delete_restitutions(
            event_annotation
        )
        deleted_restitutions = event_annotation.update_documents_after_delete_restitution(
            deleted_restitutions
        )
        self.annotations.delete_all_annotations(event_annotation)
        return deleted_restitutions

    def show_annotation_widget(self):
        """Shows the annotation widget view. 
            
        The annotation widget is a tool that allows developers 
        to annotate each of the events that make up a corpus. 
        Through the annotation widget developers can tag all 
        relevant information required for the design of their chatbot. 
            
        Returns:
            Returns the view of the annotation widget.
        """
        annotation_widget = self.annotations_widget.annotations_widget_section(
            self
        )
        return annotation_widget

    def show_annotations_validator_widget(self):
        annotations_validator_widget = self.annotations_validator_widget.get_validator_widget(
            self
        )
        return annotations_validator_widget

    def show_invalid_spans_table(self):
        invalid_spans_table = self.annotations_validator_widget.get_invalid_spans_list(
            self
        )
        return invalid_spans_table

    def show_functionals_help_widget(self):
        functionals_help_widget = self.annotations_widget.functionals_help_widget_section(
            self
        )
        return functionals_help_widget

    def show_disambiguation_widget(self):
        """Shows the disambiguation widget view.

        The disambiguation widget is a tool that allows developers
        to access all ambiguous lemma inputs.
        Ambiguity occurs when a word is susceptible to two or more
        meanings and interpretations, so the same word can be 
        understood in different ways. 
        Therefore, through the disambiguation widget developers 
        can disambiguate a lemma with multiple inputs to the lexicon.
        Developers can identify and select the appropriate meaning of a lemma 
        within the corpus and lexicon.
            
        Developers can disambiguate:
        Non-functional lemmas: nouns, adjectives, verbs and adverbs.
        Developers can select the appropriate meaning of a word within
        the corpus and lexicon. 
            
        Functional lemmas: prepositions, articles, pronouns, conjunctions. 
        Developers can choose the prototypical scheme that
        identifies the function of a lemma in a context.
        Also, developers can also create their own word classes. 
        Nouns, adjectives, articles are classes that are delivered with
        the standard, but other custom classes can be built.

        Returns:
            Returns the view of the disambiguation widget .
        """
        disambiguation_widget = self.disambiguation_widget.disambiguation_widget_section(
            self
        )
        return disambiguation_widget

    def test_paradigm(
        self,
        word_model: str,
        grammar: str,
        descriptions: Dict[str, Any],
        words: List[str],
        regional_settings: str,
        words_inflections: List = [],
    ) -> widgets.widget_box.VBox:
        """Tests a morphological paradigm in a word.

        This method allows developers to test a morphological paradigm
        of a word.
        Through morphological paradigms developers can see the inflections
        of a specific word.
        The morphological paradigms are constructed according to the following
        syntax:
        - Commands and characters can be combined in any way.
        - The FEATURES that you want to associate to a rule are combined with
            the '+' sign and separated from the commands with a '/'.
        - A paradigm can contain many concatenated rules using the '|' as a
            separator.
        - The created paradigms can be reused to create new paradigms in the
            same excel sheet or in some other simply by referencing them like
            this: <sheet:word_model>.

        Commands:
        <E>: Empty element. The word is not modified in any way.
        <L[#]>: Pointer.  One or more characters to the left.
        <R[#]>: Pointer.  One or more characters to the right.
        <A>: Deletes the check mark from the character.
        <B[#]>: Delete one or more characters and move the pointer to the left.
        <Á>: Adds an acute accent to the pointed vowel.
        <Ä>: Adds dieresis/trema/umlaut to the pointed vowel.

        Args:
            word_model (str): is a word that allows to identify the paradigm.
            grammar (str): morphological rule in the form of grammar following
                the syntax explained above.
            descriptions (Dict[str, Any]): a description for each kind of
                feature specified in the grammar.
            words: contains a set of words on which the rule in grammar
                will be tested.
            regional_settings: specifies the language of the input.
                Use 'es' for Spanish or 'en' for English.
            words_inflections (List[str]): contains a list of the inflections
                of a word.
                
        Returns:
            Returns a list table view of inflected words.
        """
        res = []
        producer_group = ParadigmProducerGroup(
            "", self.standard, regional_settings
        )
        producer = ParadigmProducer(
            word_model, grammar, False, descriptions, producer_group
        )
        for i in range(0, len(words)):
            ms = producer.apply(words[i])
            wi = []
            if len(words_inflections) != 0:
                wi = words_inflections[i]
            res.append(widgets_utilities.show_paradigm(ms, wi))
        return widgets.VBox(res)

    def test_paradigm_excel(
        self,
        pos_id: str,
        word_model: str,
        words,
        regional_settings,
        words_inflections=[],
    ):
        res = []
        for i in range(0, len(words)):
            ms = self.standard.get_morphemes(
                pos_id, word_model, words[i], regional_settings
            )
            wi = []
            if len(words_inflections) != 0:
                wi = words_inflections[i]
            res.append(widgets_utilities.show_paradigm(ms, wi))
        return widgets.VBox(res)

    def create_conjugator_section(
        self, verb: str, language_tool: str = "es-wr"
    ):
        """Creates a verb conjugation section.

        This method allows the developer to check the conjugation of a verb 
        in Spanish according to wordreference.com and in English according
        to verbix.com website.

        Args:
            verb (str): verb to be conjugated.
            language_tool (str): language of the input.
                Use 'es-wr' for Spanish (Word Reference website) or
                'en' for English (Verbix website).

        Returns:
            Creates a conjugation section and returns the conjugation of a
                verb in the language selected.
        """
        res = widgets_utilities.create_conjugator_section(verb, language_tool)
        return res

    def create_rae_conjugation_section(self):
        """The method shows the RAE models of verbal conjugation. 

        Through this method the developers can see some tables
        that show the model for the conjugation of regular and irregular
        verbs.
        
        Returns:
             Returns the view of the RAE conjugation models.
        """
        res = widgets_utilities.create_rae_conjugation_section()
        return res

    def clean_word_reference_conjugation(self, list_str: str) -> List[str]:
        """Cleans Word Reference conjugator section.

        The 'clean word reference conjugation' method allows you get a list
        of verb inflections from the conjugation section of Word Reference
        website. This method cleans or erases unnecessary characters in order
        to get a "clean list" able to be used in any purpose need it.

        Example:
            >>> input = '''I read
            >>> you read
            >>> he
            >>> she
            >>> it reads
            >>> we read
            >>> you read
            >>> they read'''
            >>>
            >>> self.creat_conjugation_section(input)
            ['read', 'read', 'reads', 'read', 'read', 'read', 'read']

        Args:
            list_str (str): list of inflections of a verb from the conjugation
                section of Word Reference website.

        Returns:
            Returns a clean list of the inflections of a verb.
        """
        list_split = list_str.split("\n")
        list_res = []
        for c in list_split:
            i = c.find("\t")
            c_n = c[i + 1 :] if i > 0 else c
            list_res.append(c_n)
            print(c_n)
        return list_res

    def get_ready_lemmas_table(
        self,
        regional_settings: str,
        layer: int = 0,
        as_data_frame: bool = True,
    ):
        """Gets ready lemmas table from.
        
        The 'ready lemmas table' is a plug-in to the lexicalization widget.
        This table allows developers to see the lemmas that are part of
        the events in the corpus and that have been lexicalized.
        This table includes lexicalized lemmas coming directly from
        corpus events (level 1) as well as lemmas coming from definitions
        created by the developer in previous lexicalization
        stages (level 2 or higher).

        Ready lemmas can be one of:
            - lemmas that has been lexicalized
            - functional lemmas
            - semantic primes
            - synonyms

        Args:
            regional_settings: specifies the language of the lexicon. 
                Use 'es' for Spanish or 'en' for English.
            layer: specifies the level of the lexicon. Zero-based. 
            as_data_frame: specifies the format of the return
                to be a pandas Dataframe. If False, Renders the widget_string
                value as HTML.
                
        Returns:
            Returns the view of a table of already lexicalized lemmas to
            the level indicated in the layer argument.
        """
        return widgets_utilities.create_ready_words_section(
            self.dictionary, regional_settings, layer, as_data_frame
        )

    def get_pending_words_table(
        self,
        regional_settings: str,
        layer: int = 0,
        as_data_frame: bool = True,
    ):
        """Gets pending words table.

        In the process of creation of a lexicon or internal dictionary, 
        the 'pending words table' allows the developer to identify words
        that have not been lexicalized. 'Pending words' are words that come
        from the events of a corpus (level 1) or from definitions created by
        the developer in previous stages of lexicalization (level 2 or higher)

        Args:
            regional_settings (str): specifies the language of the lexicon.
                Use 'es' for Spanish or 'en' for English.
            layer (int): specifies the level of the lexicon. Zero-based.
            as_data_frame (bool): specifies the format of the return to be a
                Pandas' DataFrame. If False, renders the widget_string value
                as HTML.

        Returns:
            Returns the view of a table with words that have not been
            lexicalized at the specified level.
        """
        return widgets_utilities.create_pending_words_section(
            self.dictionary, regional_settings, layer, as_data_frame
        )

    def get_undefined_functionals_table(
        self, regional_settings: str, as_data_frame: bool = True
    ):
        """Gets undefined functional table.

        The 'undefined functionals table' component is a plug-in to the
        lexicalization widget. Through this add-on developers can see all
        functional words that are part of the corpus to be lexicalized.
        The words shown in this table are functional words that have not been
        modeled through Mammut's cognitive models.

        These models organize and analyze the functionality of words in
        discourse. Each of the words that are part of the events of the corpus
        belongs to a category:

        Non-functional words are those that can be defined through the
        lexicalization widget. Example: Nouns, adjectives, verbs and adverbs.

        Functional words are those that can be defined through the
        lexicalization widget. Example: Pronouns, articles, prepositions, and
        conjunctions. These functional words are treated separately in
        the lexicon.
        
        Args:
            regional_settings: specifies the language of the lexicon. 
                Use 'es' for Spanish or 'en' for English.
            as_data_frame: specifies the format of the return
                to be a pandas' Dataframe. If False, Renders the widget_string
                value as HTML.

        Returns:
            Returns the view of a table with the functional words not defined
            in the lexicon
        """
        return widgets_utilities.create_undefined_functionals_section(
            self.standard, regional_settings, as_data_frame
        )

    def get_all_ready_lemmas_table(
        self, regional_settings: str, as_data_frame: bool = True
    ):
        """Gets all ready lemmas table.

        The 'all ready lemmas table' is a plug-in to the lexicalization
        widget. This table allows developers to see all the lemmas that are
        part of the events in the corpus and that have been lexicalized.

        This table includes lexicalized lemmas coming directly from corpus
        events (level 1) as well as lemmas coming from definitions created
        by the developer in previous lexicalization stages (level 2 or higher)

        All ready lemmas can be one of:
            - lemmas that has been lexicalized
            - functional lemmas
            - semantic primes
            - synonyms

        Args:
            regional_settings: specifies the language of the lexicon. 
                Use 'es' for Spanish or 'en' for English.
            as_data_frame: specifies the format of the return
                to be a pandas' Dataframe. If False, Renders the widget_string
                value as HTML.

        Returns:
            Returns the view of a table with all the words that have been
            lexicalized.
        """
        return widgets_utilities.create_ready_words_section(
            self.dictionary, regional_settings, sys.maxsize, as_data_frame
        )

    def get_all_pending_words_table(
        self, regional_settings: str, as_data_frame: bool = True
    ):
        """This method gets all pending words table.

        In the process of creation of a lexicon or internal dictionary, 
        the 'all pending words table' allows the developer to identify all
        words that have not been lexicalized. 'Pending words' are words that
        come from the events of a corpus (level 1) or from definitions created
        by the developer in previous stages of
        lexicalization (level 2 or higher).

        Args:
            regional_settings (str): specifies the language of the lexicon.
                Use 'es' for Spanish or 'en' for English.
            as_data_frame (bool): specifies the format of the return to be a
                Pandas' DataFrame. If False, renders the widget_string value
                as HTML.

        Returns:
            Returns the view of a table with words of all levels which have
            not been lexicalized.
        """
        return widgets_utilities.create_pending_words_section(
            self.dictionary, regional_settings, sys.maxsize, as_data_frame
        )

    def get_words_in_definitions_out_of_vocabulary(
        self, regional_settings: str, as_data_frame: bool = True
    ):
        """This method allows the developer to know which words out of the
        lexicon could be lexicalized.
        
        These are words found in the lexicalized definitions
        that the developer builds through the lexicalization process.
        
        For the Framework a word is considered 'out of the vocabulary' if it
        is not present neither in the annotable corpus nor in the list of
        functional and non-functional words of the Standard, nor in the list
        of synonyms or primes.

        Args:
            regional_settings (str): specifies the language of the lexicon.
                Use 'es' for Spanish or 'en' for English.
            as_data_frame (bool): specifies the format of the return to be a
                Pandas' DataFrame. If False, renders the widget_string
                value as HTML.

        Returns:
            Returns the view of a table with all the words in the vocabulary.
        """
        return widgets_utilities.create_words_in_definitions_out_of_vocabulary_section(
            self.dictionary, regional_settings, as_data_frame
        )

    def show_lexicalization_evolution(self, regional_settings: str):
        """This method shows a line graph representing data on the evolution
        and size of the lexicon.

        This graph helps the developer to know the relationship between the
        amount of lexicalized lemmas and the amount of words that have not
        been lexicalized, in order to see how close, the lexicon is to have
        a balance.

        Args:
            regional_settings (str): specifies the language of the lexicon.
                Use 'es' for Spanish or 'en' for English.

        Returns:
            Returns the plot of the lexicalization evolution.
        """
        widgets_utilities.show_lexicalization_evolution(
            self.tokenizer, regional_settings
        )

    def get_lexicalization_statistics(self, regional_settings: str):
        """This method collects statistics of the lexicon about number of
        lemmas, number of definitions, number of morphemes, number of
        lexicalized lemmas, and number of undefined words.

        Example:
            >>> self.get_lexicalization_statistics('es')
            LexicalizationStatistics(
              lemmas_count=675,
              definitions_count=804,
              morphemes_count=11356,
              words_count=654,
              undefined_words_count=38
              )


        Args:
            regional_settings (str): specifies the language of the lexicon.
                Use 'es' for Spanish or 'en' for English.

        Returns:
            Returns lexicalization statistics. 
        """
        return self.dictionary.get_lexicalization_statistics(regional_settings)

    def get_similar_phrases(
        self, corpus_name: str, corpus_type: str, query: str
    ) -> List[str]:
        """Looks in a corpus for all phrases similar to one provided.
        
        Through this method developers can look at a corpus (N, M, C) 
        all phrases similar to a phrase provided as input. 
        This method checks the phrase provided as input and returns 
        outputs similar to the input. 
        Semantic similarity in the area of natural language processing
        is the measure of the interrelationship between any two words
        in a text.
        Two words or terms have a similar context because they exist in 
        the same document. It is understood that these two words are related,
        and therefore their semantic distance can be deduced.
        However, in this context, two phrases are considered similar 
        if the output phrases have a meaning close to the input phrase. 
        Two phrases are similar if one can be used instead of the other.
        
        Example: 
            >>> package.get_similar_phrases(
            >>> 'corpus_degusta',
            >>> CorpusMap.DEFAULT_CRAWLER_TYPE,
            >>> 'un buen lugar para almorzar'
            >>> )
            
            [
            'Que buen local,la atencion es muy buena y el ambiente excelente',
            'El mejor ambiente para cenar. Totalmente recomendado.',
            'encantada con este sitio, vale la pena darse el gusto.',
            'Comidas deliciosas y con un ambiente muy calido.',
            'Variedad de platos de excelente calidad recomendado.',
            'Me parece muy buen restaurante aunque un poco costoso.',
            'Excelente atención, y un menú de primera',
            'Definitivamente uno de los mejores lugares para comer sushi',
            'Muy bueno de verdad muy buena opción en la ciudad'
            ]

        Args:
            corpus_name (str): name of the corpus to check.
            corpus_type (str): type of the corpus to check.
            query (str): phrase which we are looking similarities for.

        Returns:
            Returns a list of similar phrases found in the corpus specified.
            Specifically, the method returns 10 output sentences similar
            to input query sentence.
        """
        if self.paraphrases is None:
            from mammut.linguistics.paraphrases import Paraphrases

            self.paraphrases = Paraphrases(self)

        return self.paraphrases.get_similar_phrases(
            corpus_name, corpus_type, query
        )

    # Templates temporal

    def span_calculator(
        self, dataframe: pd.DataFrame, source: List[str], target: List[str]
    ):
        """Span Calculator

        This method takes a source list with the variables to substitute in
        the annotations and a target list with the words to be included in
        the annotations. Then, shows a Dataframe that has the annotation of
        a single sentence, on a single level.
        
        Args:
            dataframe: Pandas' object that has the annotation of a single
                sentence, on a single level
            source(list): list with the variables to substitute in the
                annotations.
            target(list): list with the words to be included in the
                annotations

        Returns:
            Shows the Dataframe with the changes.
        """
        span_series = []
        original_span_series = []

        for num_variable in range(len(source)):
            variable = source[num_variable]
            variant = target[num_variable]
            if num_variable == 0:
                for s in dataframe[dataframe.text == variable].span:
                    previous_span = (
                        s  # Mejorar esta forma de seleccionar un span
                    )
            else:
                previous_span = new_span_series[num_variable]
            growth = len(variant) - len(variable)
            try:
                start = int(previous_span[0])
            except UnboundLocalError:
                print(
                    "Error: La palabra <<{0}>> no se encuentra en el dataframe".format(
                        source[num_variable]
                    )
                )
                break

            # Obtención de los spans originales:

            if len(span_series) == 0:
                for i in dataframe.span:
                    try:
                        if str(type(i)) == """<class 'str'>""":
                            comma_pos = i.find(",")
                            span_tuple = (
                                int(i[:comma_pos]),
                                int(i[comma_pos + 1 :]),
                            )
                            span_series.append(span_tuple)
                            original_span_series.append(span_tuple)
                        else:
                            span_series.append("nan")
                    except:
                        span_series.append("nan")

            new_span_series = []
            for j in span_series:
                if str(type(j)) == """<class 'tuple'>""":
                    if j[0] == start:
                        new_span_tuple = (start, j[1] + growth)
                    elif j[0] > start:
                        new_span_tuple = (j[0] + growth, j[1] + growth)
                    elif j[0] < start and j[1] <= start:
                        new_span_tuple = (j[0], j[1])
                    else:
                        new_span_tuple = (j[0], j[1] + growth)
                    new_span_series.append(new_span_tuple)
                else:
                    new_span_series.append("nan")
            span_series = new_span_series

            dataframe = dataframe.replace({variable: variant}, regex=True)

        # Remplazar los spans en la columna span:

        for span in range(len(new_span_series)):
            if str(type(new_span_series[span])) == """<class 'tuple'>""":
                dataframe.iloc[span, 2] = str(new_span_series[span])[
                    1:-1
                ].replace(" ", "")
            else:
                continue

        # Remplazar los spans en otras columnas:

        corpus_name = (
            dataframe.iloc[0, 0][: dataframe.iloc[0, 0].find("SP")] + "SP-"
        )
        for n in range(len(original_span_series)):
            if str(type(original_span_series[n])) == """<class 'tuple'>""":
                first_number, second_number = (
                    original_span_series[n][0],
                    original_span_series[n][1],
                )
                id_regex = re.compile(corpus_name + r"(\d){1,4}-0-0")
                sentence_num = id_regex.search(dataframe.iloc[0, 0]).group(1)
                source_id = "{3}{0}-0-0-{1}-{2}-".format(
                    sentence_num, first_number, second_number, corpus_name
                )
                target_id = "{3}{0}-0-0-{1}-{2}-".format(
                    sentence_num,
                    str(new_span_series[n][0]),
                    str(new_span_series[n][1]),
                    corpus_name,
                )
                dataframe = dataframe.replace(
                    {source_id: target_id}, regex=True
                )
            else:
                continue

        return dataframe

    def id_replace(self, dataframe: pd.DataFrame, id_number: int):
        """ID Replace

        This method accepts as a second argument a number that will indicate 
        the number of the annotated sentence.

        Args:
            dataframe: Pandas' object that has the annotation of a single
                sentence, on a single level.
            id_number(int): number that will indicate the number of the
                annotated sentence.
        
        Returns:
            Shows the Dataframe.
        """
        corpus_name = (
            dataframe.iloc[0, 0][: dataframe.iloc[0, 0].find("SP")] + "SP-"
        )
        id_regex = re.compile(corpus_name + r"(\d+)-0-0")
        sentence_num = id_regex.search(dataframe.iloc[0, 0]).group(1)
        source_id = "{0}{1}-0-0".format(corpus_name, sentence_num)
        print(source_id)
        target_id = "{0}{1}-0-0".format(corpus_name, str(id_number))
        dataframe = dataframe.replace({source_id: target_id}, regex=True)
        return dataframe

    def build_synthetic_corpus(
        self,
        nouns_words: int,
        adjective_words: int,
        verbs_words: int,
        nouns_bound: int,
        adjectives_bound: int,
        verbs_bound: int,
        free_lemmas_word: int,
        random_ext: int
    ):
        """Returns a synthetic corpus using the provided guidelines

        Args:
            nouns_words(int): number of times that a noun bound morpheme
                will appear in company of a random free morpheme.
            adjective_words(int): number of times that an adjective bound
                morpheme will appear in company of a random free morpheme.
            verbs_words(int): number of times that a verb bound morpheme
                will appear in company of a random free morpheme.
            nouns_bound(int): number of times that a noun bound morpheme
                will appear alone in the corpus.
            adjectives_bound(int): number of times that an adjective
                bound morpheme will appear alone in the corpus.
            verbs_bound(int): number of times that a verb bound morpheme
                will appear alone in the corpus.
            free_lemmas_word(int): number of times that a free morpheme
                will appear in the corpus with a random suffix.

        Returns
            A list of strings. Each string is a word in the synthetic
            corpus.
        """
        return self.synthetic_corpus_generator.generate_synthetic_corpus(
            nouns_words,
            adjective_words,
            verbs_words,
            nouns_bound,
            adjectives_bound,
            verbs_bound,
            free_lemmas_word,
            random_ext
        )

    def get_all_bound_morphemes(self, regional_settings: str) -> List[str]:
        """Retrieve all the bounded morphemes founded in package.
        All the bound morphemes are retrieved for all POS and
        for the specified regional_settings.

        Todo: Only 'es' regional setting is supported.
        Todo: Only POS nouns, adjectives and some verbs models are
            supported

        Each returned element will be a string composed of the
        free morpheme in a randomly altered form and the bound
        morpheme.
        Args:
            regional_settings(str): language of the bound morphemes
                models
        Returns:
            A list containing the words composed of the altered
            free morpheme and the bound morpheme.
        """
        ret = self.morphemes_generator.get_tokens("NOUN", regional_settings)
        ret = ret + self.morphemes_generator.get_tokens(
            "ADJ", regional_settings
        )
        ret = ret + self.morphemes_generator.get_tokens(
            "VERB", regional_settings
        )
        return ret

    def compile_corpus_list(
        self,
        mammut_id: int,
        compilation_message_callback,
        corpus_ids: List[str],
    ):
        """Used to compile a Mammut bot.

        Args:
            mammut_id(int): the bot user ID in the API database.
            compilation_message_callback: callback to update the simulator
                widget output.
            corpus_ids(List[str]): list of corpus IDs to compile according
                to the corpus_map sheet.

        """
        if self.main_id is not None:
            if self.corpus_compiler_service == None:
                self.corpus_compiler_service = CorpusCompilerService()
            return self.corpus_compiler_service.compile(
                mammut_id,
                self.main_id,
                True,
                "en",
                "->",
                corpus_ids,
                compilation_message_callback,
            )
        else:
            compilation_message_callback(
                "Error: para compilar se debe especificar "
                "un presentation id como paquete",
                None,
            )

    @staticmethod
    def image_retraining(
        self, directory: str, mammut_id: int, samples: int = 50,
    ):
        # TODO:Debemos separar todo lo relacionado con vision
        # Para asi no forzar que siempre se deba instalar ciertas librerias.
        from mammut.models.imaging.mammut_image_retraining import (
            augmentate_and_retrain,
        )

        return augmentate_and_retrain(directory, mammut_id, samples,)

    def get_image_retraining_widget(self, mammut_user):
        # TODO:Debemos separar todo lo relacionado con vision
        # Para asi no forzar que siempre se deba instalar ciertas librerias.
        from mammut.visualization.image_retraining_widget import (
            get_image_retraining_widget,
        )

        return get_image_retraining_widget(self, mammut_user)

    @staticmethod
    def validate_challenge(class_name: str,
                           public_field_name: str,
                           private_field_name: str,
                           property_name: str,
                           public_instance_method_name: str,
                           public_static_method_name: str):
        """Hello, I'll help you to validate your challenge.

        For this challenge you shall write a Class. Whatever class you want,
        but ensure that your class has:
            - A public field name
            - A private field name
            - A property name
            - An instance method name
            - A class method (static) name

        If your class has constructor parameters, ensure that they have a default
        value.

        I need the following args:

        Args:
            class_name(str): The name of the class that you've defined above.
            public_field_name(str): The name of the public field
            private_field_name(str): The name of the private field
            property_name(str): The name of the property
            public_instance_method_name(str): The name of the instance method.
            public_static_method_name(str): The name of the class method.

        Returns:
            "Success, you've made it!" if correct
            "Test failed" if failed. And maybe some hint :)
        """
        import sys
        from types import FunctionType
        obj = getattr(sys.modules['__main__'], class_name)

        # Test class method
        try:
            static_mth = obj.__dict__[public_static_method_name]
            if not isinstance(static_mth, staticmethod) and not isinstance(static_mth, classmethod):
                return f"Test failed: Class method {public_static_method_name} not found :("
        except:
            return f"Test failed: Class method {public_static_method_name} not found :("

        # Test instance method
        try:
            instance_mth = obj.__dict__[public_instance_method_name]
            if not isinstance(instance_mth, FunctionType):
                return f"Test failed: Instance method {public_instance_method_name} not found :("
        except:
            return f"Test failed: Instance method {public_instance_method_name} not found :("

        # Test class property
        try:
            prop = obj.__dict__[property_name]
            if not isinstance(prop, property):
                return f"Test failed. Property {property_name} not found :("
        except:
            return f"Test failed. Property {property_name} not found :("

        try:
            instance = obj()
        except:
            return f"Test failed. Does your class has default values for constructor parameters?"

        fields = vars(instance)
        if not public_field_name in fields and not public_field_name in obj.__dict__:
            return f"Test failed. Public field {public_field_name} not found :("

        if not private_field_name in fields and not private_field_name in obj.__dict__:
            return f"Test failed. Private field {private_field_name} not found :("

        return "Success, you've made it!"








# coding=utf-8
import datetime
import functools
from enum import Enum, auto
from mammut.common.storage.storage_manager import StorageManager
import requests
import json
import mammut
from functools import reduce


from mammut.common.util import StringEnum, RegionalSettingsType


class MicroExperience(StringEnum):
    call = "call"
    share = "share"
    web = "web"
    app = "app"
    postback = "postback"
    quickreply = "quickreply"


class WebviewSize(StringEnum):
    compact = "compact"
    tall = "tall"
    full = "full"


class UserType(StringEnum):
    supervisor = "supervisor"
    person = "person"
    normal = "normal"
    machine = "machine"
    facebook = "facebook-automatic"


class VerificationStatus(StringEnum):
    unverified = "unverified"
    verified = "verified"
    unknown = "unknown"


class Gender(StringEnum):
    male = "male"
    female = "female"
    other = "other"


class RoomType(StringEnum):
    group = "group"
    private = "private"


class PosiblePlacesFormat(StringEnum):
    textField = 0
    select = 1
    checkBox = 2
    video = 3
    image = 4
    url = 5


class CountryCodes(StringEnum):
    VE = "+58"
    CH = "+56"
    US = "+1"


class User:
    def __init__(
        self,
        first_name: str,
        last_name,
        nickname,
        main_phone_number: str,
        main_email: str,
        locale: RegionalSettingsType,
        gender: Gender,
        birthdate: datetime.date,
        verification_status: VerificationStatus,
        user_type: UserType,
        facebook_id: str = "",
        facebook_access_token: str = "",
        rooms: list = [],
        bot_name: str = "",
    ):
        self.first_name = first_name
        self.last_name = last_name
        self.nickname = nickname
        self.main_phone_number = main_phone_number
        self.main_email = main_email
        self.locale = locale
        self.gender = gender
        self.birthdate = birthdate
        self.verification_status = verification_status
        self.user_type = user_type
        self.facebook_id = facebook_id
        self.facebook_access_token = facebook_access_token
        self.verbs = {}
        self.rooms = rooms
        self.bot_name = bot_name
        self.id = None

    @classmethod
    def from_json(cls, api_user_json: {}):
        first_name = ""
        last_name = ""
        nickname = ""

        # Get names from proper noun structures

        if "name" in api_user_json.keys():
            first_name = api_user_json["name"]
        if "lastname" in api_user_json.keys():
            last_name = api_user_json["lastname"]
        if "nickname" in api_user_json.keys():
            nickname = api_user_json["nickname"]

        id = api_user_json["id"]

        main_phone_number = ""
        if "main-phone-number" in api_user_json.keys():
            main_phone_number = api_user_json["main-phone-number"]
        main_email = ""
        if "main-email" in api_user_json.keys():
            main_email = api_user_json["main-email"]
        locale = ""
        if "locale" in api_user_json.keys():
            temp_local = api_user_json["locale"].lower()
            if temp_local == "sp":
                temp_local = "es"
            locale = RegionalSettingsType(temp_local)
        gender = ""
        if "gender" in api_user_json.keys():
            gender = Gender[api_user_json["gender"]]
        birthdate = None
        if "birthdate" in api_user_json.keys():
            birthdate = mammut.str_parse_datetime(api_user_json["birthdate"]).date()
        verification_status = None
        if "verification-status" in api_user_json.keys():
            verification_status = VerificationStatus[
                api_user_json["verification-status"]
            ]
        user_type = None
        if "user-type" in api_user_json.keys():
            user_type = UserType(api_user_json["user-type"])

        facebook_id = ""
        if "facebook-id" in api_user_json.keys():
            facebook_id = api_user_json["facebook-id"]

        facebook_access_token = ""
        if "facebook-access-token" in api_user_json.keys():
            facebook_access_token = api_user_json["facebook-access-token"]

        rooms = []
        if "chat-in" in api_user_json.keys():
            for room in api_user_json["chat-in"]["room"]:
                rooms.append(Room.from_json(room))
        elif "create" in api_user_json.keys():
            for room in api_user_json["create"]["room"]:
                rooms.append(Room.from_json(room))

        bot_name = ""
        if "bot-name" in api_user_json.keys():
            bot_name = api_user_json["bot-name"]

        user = User(
            first_name,
            last_name,
            nickname,
            main_phone_number,
            main_email,
            locale,
            gender,
            birthdate,
            verification_status,
            user_type,
            facebook_id,
            facebook_access_token,
            rooms,
            bot_name,
        )
        user.id = id
        return user

    def to_json(self):
        data = {}
        if self.main_phone_number != "":
            data["main-phone-number"] = self.main_phone_number
        data["main-email"] = self.main_email
        if self.locale is not "":
            data["locale"] = self.locale.value
        if self.gender is not "":
            data["gender"] = self.gender.value
        if self.birthdate is not None:
            data["birthdate"] = self.birthdate.strftime(mammut.LOCAL_DATE_FORMATTER)
        if self.verification_status is not None:
            data["verification-status"] = self.verification_status.value
        if self.user_type is not None:
            data["user-type"] = self.user_type.value
        if self.facebook_id != "":
            data["facebook-id"] = self.facebook_id
        if self.facebook_access_token != "":
            data["facebook-access-token"] = self.facebook_access_token
        data["name"] = self.first_name
        if self.last_name is not None:
            data["lastname"] = self.last_name
        if self.bot_name is not "":
            data["bot-name"] = self.bot_name

        if self.id is not None:
            data["id"] = self.id

        return data

    def add_verb(
        self,
        scoped: bool,
        id: int,
        title: str,
        micro_experience: str,
        phone_number: str,
        url: str,
        url_parameters: str,
        payload: str,
        image: str,
        webview_size: str,
        knowledge_graph,
    ):
        verb = Verb(
            self,
            scoped,
            id,
            title,
            micro_experience,
            phone_number,
            url,
            url_parameters,
            payload,
            image,
            webview_size,
            knowledge_graph,
        )
        self.verbs[id] = verb
        return verb

    def has_verb(self, id: int):
        return id in self.verbs

    def get_verb(self, id: str):
        return self.verbs[id]


class VerbInstance:
    def __init__(
        self,
        verb,
        title: str,
        micro_experience: MicroExperience,
        phone_number: str = None,
        url: str = None,
        url_parameters: str = None,
        payload: str = None,
        image: str = None,
        webview_size: str = None,
        vertex_title: str = None,
        vertex_index: str = None,
    ):
        self.verb = verb
        self.title = title
        self.micro_experience = micro_experience
        self.url = url
        if url_parameters:
            if isinstance(url_parameters, list):
                self.url = self.url.format(*url_parameters)
            else:
                self.url = self.url.format(url_parameters)
        self.phone_number = phone_number
        self.payload = payload
        self.image = image
        self.webview_size = webview_size
        self.api_id = 0
        self.vertex_title = vertex_title
        self.vertex_index = vertex_index


class VerbDefinitionType(Enum):
    Singleton = auto()
    OneInstancePerMainVertex = auto()
    MultipleInstancesPerMainVertex = auto()


class Verb:
    def __init__(
        self,
        user: User,
        scoped: bool,
        id: int,
        title: str,
        str_micro_experience: str,
        phone_number: str,
        url: str,
        url_parameters: str,
        payload: str,
        image: str,
        webview_size: str,
        knowledge_graph,
    ):
        def add_error_of_invalid_button(error_msg, invalid_tokens):
            self.valid = False
            invalid_tokens_str = functools.reduce(
                lambda l, r: l + r, invalid_tokens, ""
            )
            error_msg = "Invalid button: " + error_msg + f". Id = {self.id}"
            if len(invalid_tokens) > 0:
                error_msg = error_msg + f". Invalid tokens: {invalid_tokens_str}"
            self.errors.append(error_msg)

        self.valid = True
        self.errors = []
        self.user = user
        self.scoped = scoped
        self.id = id
        self.title = title
        self.str_micro_experience = str_micro_experience
        self.url = url
        self.url_parameters = url_parameters
        self.phone_number = phone_number
        self.payload = payload
        self.image = image
        self.webview_size_str = webview_size
        self.webview_size = None
        self.knowledge_graph = knowledge_graph
        self.definition_type = VerbDefinitionType.Singleton
        self.singleton_instance = None
        self.instances = {}

        if self.str_micro_experience in list(MicroExperience.__members__):
            self.micro_experience = MicroExperience(self.str_micro_experience)
        else:
            self.micro_experience = MicroExperience.share
            add_error_of_invalid_button("has invalid micro-experience", [])
        self.tokenized_title = self.knowledge_graph.tokenizer.tokenize_corpus_entry(
            self.title, False, self.user.locale.value, self.knowledge_graph
        )
        self.has_variable_title = len(self.tokenized_title.variables) > 0
        if len(self.tokenized_title.invalid_variables) > 0:
            add_error_of_invalid_button(
                "has a invalid variable in the title",
                self.tokenized_title.invalid_variables,
            )
        if self.micro_experience == MicroExperience.call:
            self.tokenized_phone_number = self.knowledge_graph.tokenizer.tokenize_corpus_entry(
                self.phone_number, False, self.user.locale.value, self.knowledge_graph
            )
            self.has_variable_phone_number = (
                len(self.tokenized_phone_number.variables) > 0
            )
            if self.tokenized_phone_number.contains_invalid_variables():
                add_error_of_invalid_button(
                    "has a invalid variable in the phone number",
                    self.tokenized_phone_number.invalid_variables,
                )
        elif (
            self.micro_experience == MicroExperience.web
            or self.micro_experience == MicroExperience.app
        ):
            self.tokenized_url = self.knowledge_graph.tokenizer.tokenize_corpus_entry(
                self.url, False, self.user.locale.value, self.knowledge_graph
            )
            self.has_variable_url = len(self.tokenized_url.variables) > 0
            if self.tokenized_url.contains_invalid_variables():
                add_error_of_invalid_button(
                    "has a invalid variable in the url",
                    self.tokenized_url.invalid_variables,
                )
            self.tokenized_url_parameters = self.knowledge_graph.tokenizer.tokenize_corpus_entry(
                self.url_parameters, False, self.user.locale.value, self.knowledge_graph
            )
            self.has_variable_url_parameters = (
                len(self.tokenized_url_parameters.variables) > 0
            )
            if self.tokenized_url_parameters.contains_invalid_variables():
                add_error_of_invalid_button(
                    "has a invalid variable in the url parameters",
                    self.tokenized_url_parameters.invalid_variables,
                )

            if self.webview_size_str in list(WebviewSize.__members__):
                self.webview_size = WebviewSize(self.webview_size_str)

        elif (
            self.micro_experience == MicroExperience.postback
            or self.micro_experience == MicroExperience.quickreply
        ):
            self.tokenized_payload = self.knowledge_graph.tokenizer.tokenize_corpus_entry(
                self.payload, False, self.user.locale.value, self.knowledge_graph
            )
            self.has_variable_payload = len(self.tokenized_payload.variables) > 0
            if self.tokenized_payload.contains_invalid_variables():
                add_error_of_invalid_button(
                    "has a invalid variable in the payload",
                    self.tokenized_payload.invalid_variables,
                )
            if self.micro_experience == MicroExperience.quickreply:
                self.tokenized_image = self.knowledge_graph.tokenizer.tokenize_corpus_entry(
                    self.image, False, self.user.locale.value, self.knowledge_graph
                )
                self.has_variable_image = len(self.tokenized_image.variables) > 0
                if self.tokenized_image.contains_invalid_variables():
                    add_error_of_invalid_button(
                        "has a invalid variable in the image",
                        self.tokenized_image.invalid_variables,
                    )


class Issue:
    def __init__(
        self,
        room_id,
        custom_code=None,
        caused_by=None,
        issue_type=None,
        machine_id=None,
        id=None,
        causative_connected=None,
        resolution_text=None,
        causative_id=None,
        issue_state=None,
    ):
        self.room_id = room_id
        self.custom_code = custom_code
        self.caused_by = caused_by
        self.issue_type = issue_type
        self.machine_id = machine_id
        self.id = id
        self.causative_connected = causative_connected
        self.resolution_text = resolution_text
        self.causative_id = causative_id
        self.issue_state = issue_state

    def to_json(self):
        data = {}
        data["creation-date"] = "now"
        data["issue-state"] = self.issue_state
        data["resolution-text"] = self.resolution_text
        data["issue-type"] = self.issue_type
        data["custom-code"] = self.custom_code
        data["machine-id"] = self.machine_id
        data["room-id"] = self.room_id
        data["causative-connected"] = self.causative_connected
        return data

    @classmethod
    def from_json(cls, issue_json: {}):
        room_id = None
        if "room-id" in issue_json.keys():
            room_id = issue_json["room-id"]
        custom_code = None
        if "custom-code" in issue_json.keys():
            custom_code = issue_json["custom-code"]
        caused_by = None
        if "caused-by" in issue_json.keys():
            caused_by = issue_json["caused-by"]
        issue_type = None
        if "issue-type" in issue_json.keys():
            issue_type = issue_json["issue-type"]
        machine_id = None
        if "machine-id" in issue_json.keys():
            machine_id = issue_json["machine-id"]
        id = None
        if "id" in issue_json.keys():
            id = issue_json["id"]
        causative_connected = None
        if "causative-connected" in issue_json.keys():
            causative_connected = issue_json["causative-connected"]
        resolution_text = None
        if "resolution-text" in issue_json.keys():
            resolution_text = issue_json["resolution-text"]
        causative_id = None
        if "causative-id" in issue_json.keys():
            causative_id = issue_json["causative-id"]
        issue_state = None
        if "issue-state" in issue_json.keys():
            issue_state = issue_json["issue-state"]
        issue = Issue(
            room_id,
            custom_code,
            caused_by,
            issue_type,
            machine_id,
            id,
            causative_connected,
            resolution_text,
            causative_id,
            issue_state,
        )
        return issue


class Thought:
    def __init__(self, user_id, room_id, text, creation_date, issue=None):
        self.user_id = user_id
        self.room_id = room_id
        self.text = text
        self.creation_date = creation_date
        self.issue = issue

    def to_json(self):
        data = {}
        data["text"] = self.text
        data["creation-date"] = self.creation_date
        data["in@room"] = {"id": self.room_id}
        return data

    @classmethod
    def from_json(cls, api_thought_json: {}):

        room_id = api_thought_json["in"]["room"][0]["id"]
        text = api_thought_json["text"]
        creation_date = api_thought_json["creation-date"]
        user = User.from_json(api_thought_json["send-by"]["user"][0])
        issue = None
        if "cause" in api_thought_json.keys():
            issue = Issue.from_json(api_thought_json["cause"]["issue"][0])
            print("isuee from thougth", Issue.__dict__)

        thought = Thought(user.id, room_id, text, creation_date, issue)

        return thought


class Room:
    def __init__(
        self,
        title,
        room_type: RoomType,
        creation_date,
        created_by: User = None,
        url="",
        custom_code="",
        room_id=None,
        user_members: list = [],
    ):
        self.title = title
        self.room_type = room_type
        self.creation_date = creation_date
        self.id = room_id
        self.created_by = created_by
        self.members = user_members
        self.url = url
        self.custom_code = custom_code

    @classmethod
    def from_json(cls, api_room_json: {}):
        room_id = api_room_json["id"]

        users = []
        if "has" in api_room_json.keys():
            for user_member in api_room_json["has"]["user"]:
                users.append(User.from_json(user_member))
        creation_date = ""
        if "creation-date" in api_room_json.keys():
            creation_date = api_room_json["creation-date"]
        created_by = None
        if "create-by" in api_room_json.keys():
            created_by = User.from_json(api_room_json["create-by"]["user"][0])
        title = ""
        if "title" in api_room_json.keys():
            title = api_room_json["title"]
        room_type = ""
        if "room-type" in api_room_json.keys():
            room_type = api_room_json["room-type"]
        url = ""
        if "url" in api_room_json.keys():
            url = api_room_json["url"]
        custom_code = ""
        if "custom-code" in api_room_json.keys():
            custom_code = api_room_json["custom-code"]

        return Room(
            title=title,
            room_type=room_type,
            creation_date=creation_date,
            created_by=created_by,
            url=url,
            custom_code=custom_code,
            room_id=room_id,
            user_members=users,
        )

    def to_json(self):

        data = {}
        data["room-type"] = self.room_type.value
        data["creation-date"] = self.creation_date
        data["title"] = self.title
        data["url"] = self.url
        data["custom-code"] = self.custom_code
        contain = []
        contain.append({"deleted": False})
        data["contain"] = contain
        return data


class CalendarCustomValues:
    def __init__(
        self, name: str, value: list, value_type: PosiblePlacesFormat, mandatory: bool
    ):
        self.name = name
        self.value = value
        self.value_type = value_type
        self.mandatory = mandatory

    def to_json(self):
        data = {}
        data[self.name] = {}
        data[self.name]["value"] = self.value
        data[self.name]["type"] = self.value_type
        data[self.name]["mandatory"] = self.mandatory
        return data


class Calendar:
    def __init__(
        self,
        user_id: str,
        mammut_user: User,
        title: str,
        possible_places: list,
        possible_places_format: PosiblePlacesFormat,
        possible_regional_settings: RegionalSettingsType,
        custom_fields: list,
        reminder_text_templates: dict,
        country_code: str,
    ):
        self.user_id = user_id
        self.mammut_user = mammut_user
        self.title = title
        self.possible_places = possible_places
        self.possible_places_format = possible_places_format
        self.possible_regional_settings = possible_regional_settings
        self.custom_fields = custom_fields
        self.reminder_text_templates = reminder_text_templates
        self.country_code = country_code

    def escape_curly_brakets_in_dict(self, data_dict):
        serialized = json.dumps(data_dict)
        return serialized

    def to_json(self):
        data = {}
        data["tittle"] = self.title
        data["possible-places"] = self.possible_places
        data["possible-places-format"] = self.possible_places_format
        data["custom-data-template"] = self.escape_curly_brakets_in_dict(
            self.custom_fields
        )
        data["handle-by@user"] = {"id": self.mammut_user.id}
        data["reminder-text-templates"] = self.escape_curly_brakets_in_dict(
            self.reminder_text_templates
        )
        data["possible-regional-settings"] = self.possible_regional_settings
        data["country-code"] = self.country_code
        return data


class Request:

    mammutUrl = "https://372aab09.ngrok.io"
    verb_path = "{mammutURL}/app:mammut-1/graph/user:{userId}/can"
    fixed_user_path = (
        "{mammutURL}/app:mammut-1/fixed-vertex?type-name=user&config-id=mammut&depth=2"
    )
    post_user_path = "{mammutURL}/app:mammut-1/graph/user"
    post_room_path = "{mammutURL}/app:mammut-1/graph/user:{userId}/create"
    get_user_path = "{mammutURL}/app:mammut-1/graph:2/user:{userId}"
    get_user_by_mail_path = "{mammutURL}/app:mammut-1/unique-vertex?type-name=user&property-name=main-email&property-value={userMail}&depth=2"
    get_room_path = "{mammutURL}/app:mammut-1/graph:2/room:{roomId}"
    post_thought_path = "{mammutURL}/app:mammut-1/graph/user:{userId}/send:=thought"
    post_issue_update_path = '{mammutURL}/app:mammut-1/graph/issue:{issueId}?resolution-text="{message}";issue-state="close";resolved-by@user:{"id":{supervisorId}} '
    post_issue_remove_halted_path = "{mammutURL}/app:mammut-1/graph/issue"
    get_last_thought_path = "{mammutURL}/app:mammut-1/graph:2/user:{userId}/send::thought?creation-date.last(1)"
    get_thoughts_path = "{mammutURL}/app:mammut-1/graph:2/user:{userId}/send::thought?creation-date.last(10)"
    get_room_thoughts_path = "{mammutURL}/app:mammut-1/graph:2/room:{roomId}/with::thought?creation-date.last(10)"
    get_room_thoughts_path_since = '{mammutURL}/app:mammut-1/graph:2/room:{roomId}/with::thought?creation-date.since("{since}")'
    post_video_path = "{mammutURL}/app:mammut-1/graph/video"
    post_calendar_path = "{mammutURL}/app:mammut-1/graph/user:{userId}/own:=calendar"
    get_videos_path = "{mammutURL}/app:mammut-1/graph/user:{userId}/own::video"
    post_compilation_path = "{mammutURL}/app:mammut-1/actions/compile"
    get_vertex_by_unique_property_path = "{mammutURL}/app:mammut-1/unique-vertex?type-name={vertex}&property-name={prop_name}&property-value={prop_value}&depth=1"

    def __init__(self, mammut_url: str):
        self.mammutUrl = mammut_url

    def post_user(self, user: User = None, payload=None):

        if payload == None:
            data = user.to_json()
        else:
            data = payload
        response = self.post_request(self.post_user_path, data)

        if response is not None:
            # Get user id
            response_json = response.json()

            if response_json["status"] != "Failed":
                for result in response_json["taskResult"]:
                    if result["affectedElementName"] == "user":
                        print(
                            "efected elemnt id on user creation",
                            result["affectedElementId"],
                        )
                        created_user_id = result["affectedElementId"]
                        if user is not None:
                            user.id = created_user_id
                        break
                return created_user_id
            else:
                print("Response status for user creation is FAILED")
                return None
        else:
            print("Error saving User.")

    def post_room(
        self, room: Room = None, payload=None, data_is_json=False, creator=None
    ):
        if payload is not None:
            data = payload
            user_id = self.get_fixed_mammut().id
        else:
            data = room.to_json()
            user_id = room.created_by.id
        if creator is not None:
            user_id = creator.id
        url = self.post_room_path.replace("{userId}", str(user_id))
        response = self.post_request(url, data, data_is_json)

        created_room_id = None
        if response is not None:
            response_json = response.json()
            print(response_json["status"])

            if response_json["status"] != "Failed":
                for result in response_json["taskResult"]:
                    if result["affectedElementName"] == "room":
                        created_room_id = result["affectedElementId"]
                        if room is not None:
                            room.id = created_room_id
                        break
        else:
            print("Error creating room.")

        if payload is not None:
            return created_room_id
        else:
            return room

    def post_thought(self, thought: Thought):
        data = thought.to_json()
        url = self.post_thought_path.replace("{userId}", str(thought.user_id))
        response = self.post_request(url, data).json()
        return response

    def post_totblt_resolution(self, issue: Issue, supervisor: User, message: str):
        data = {}
        url = (
            self.post_issue_update_path.replace("{issueId}", str(issue.id))
            .replace("{supervisorId}", str(supervisor.id))
            .replace("{message}", message)
        )
        print(url)
        response = self.post_request(url, data).json()
        return response

    def post_remove_halted_mark_issue(self, issue: Issue):
        data = issue.to_json()
        url = self.post_issue_remove_halted_path
        response = self.post_request(url, data)
        return response

    def post_video(self, video_name, video_url, user_id):
        data = {
            "title": video_name,
            "url": video_url,
            "belong-to@user": {"id": user_id},
        }
        url = self.post_video_path
        response = self.post_request(url, data)

        if response is not None:
            response_json = response.json()
            if response_json["status"] != "Failed":
                for result in response_json["taskResult"]:
                    if result["affectedElementName"] == "video":
                        print("Created video", result["affectedElementId"])
                        break

        else:
            print("Error creating Video.")
        return response

    def post_calendar(self, calendar: Calendar):
        data = calendar.to_json()
        url = self.post_calendar_path.replace("{userId}", str(calendar.user_id))
        response = self.post_request(url, data).json()
        return response

    def post_verb(self, verb_def: Verb):

        verb_instances = []

        # Get verb instances depending the type of definition
        if (
            verb_def.definition_type
            == VerbDefinitionType.MultipleInstancesPerMainVertex
        ):
            for vil in verb_def.instances.values():
                verb_instances.extend(vil)
        elif verb_def.definition_type == VerbDefinitionType.OneInstancePerMainVertex:
            verb_instances.extend(verb_def.instances.values())
        else:
            verb_instances.insert(0, verb_def.singleton_instance)

        url = self.verb_path.replace("{userId}", str(verb_def.user.id))

        # post verb instances
        for verb in verb_instances:

            data = {}
            data["title"] = verb.title
            data["micro-experience"] = verb.micro_experience.value
            if verb.micro_experience == MicroExperience.call:
                data["phone-number"] = verb.phone_number
            elif (
                verb.micro_experience == MicroExperience.web
                or verb.micro_experience == MicroExperience.app
            ):
                data["url"] = verb.url
                if verb.webview_size:
                    data["webview-size"] = verb.webview_size.value
            elif (
                verb.micro_experience == MicroExperience.postback
                or verb.micro_experience == MicroExperience.quickreply
            ):
                if verb.payload:
                    data["payload"] = verb.payload
                if verb.micro_experience == MicroExperience.quickreply:
                    data["content-type"] = "text"
                    if verb.image:
                        data["image-url"] = verb.image

            response = self.post_request(url, data)

            if response is not None:
                # Get verb id
                response_json = response.json()

                for result in response_json["taskResult"]:
                    if result["affectedElementName"] == "verb":
                        verb.api_id = result["affectedElementId"]
            else:
                print("Error saving Verb.")

    def get_user(self, user_id):

        url = self.get_user_path.replace("{userId}", str(user_id))

        response = self.get_request(url)

        if response is None:
            print("Error getting user " + str(user_id))

        else:
            response_json = response.json()

            if (
                "user" in response_json["data"].keys()
                and len(response_json["data"]["user"]) > 0
            ):

                # Get first user response
                user = User.from_json(response_json["data"]["user"][0])
                return user
            else:
                print("No user found for: " + str(user_id))

    def get_user_by_mail(self, user_mail, raw_user=False):

        url = self.get_user_by_mail_path.replace("{userMail}", str(user_mail))

        response = self.get_request(url)

        if response is None:
            print("Error getting user " + str(user_mail))

        else:
            response_json = response.json()

            if response_json["status"] != "Failed":
                if (
                    "user" in response_json["data"].keys()
                    and len(response_json["data"]["user"]) > 0
                ):

                    # Get first user response
                    if raw_user:
                        user = response_json["data"]["user"][0]
                    else:
                        user = User.from_json(response_json["data"]["user"][0])
                    return user
                else:
                    print("No user found for: " + str(user_mail))
                    return None
            else:
                print("No user found for: " + str(user_mail))
                return None

    def get_last_sent_thought(self, user_id):

        thought = None
        url = self.get_last_thought_path.replace("{userId}", str(user_id))

        response = self.get_request(url)

        if response is None:
            print("Error getting user last thought: " + str(user_id))
        else:
            response_json = response.json()
            if (
                "thought" in response_json["data"].keys()
                and len(response_json["data"]["thought"]) > 0
            ):
                thought = Thought.from_json(response_json["data"]["thought"][0])
            else:
                print("No thought sent for: " + str(user_id))

        return thought

    def get_thoughts(self, id, from_type="User"):

        thoughts = list()
        if from_type == "User":
            url = self.get_thoughts_path.replace("{userId}", str(id))
        elif from_type == "Room":
            url = self.get_room_thoughts_path.replace("{roomId}", str(id))
            print(url)

        response = self.get_request(url)

        if response is None:
            print("Error getting user last thought: " + str(id))
        else:
            response_json = response.json()
            if (
                "thought" in response_json["data"].keys()
                and len(response_json["data"]["thought"]) > 0
            ):
                for thought_response in response_json["data"]["thought"]:
                    thoughts.append(Thought.from_json(thought_response))
            else:
                print("No thought sent for " + from_type + ": " + str(id))

        return thoughts

    def get_thoughts_since(self, id, since_date):
        thoughts = list()
        url = self.get_room_thoughts_path_since.replace("{roomId}", str(id))
        url = url.replace("{since}", since_date)
        response = self.get_request(url)
        if response is None:
            print("Error getting user last thought: " + str(id))
        else:
            response_json = response.json()
            if (
                "thought" in response_json["data"].keys()
                and len(response_json["data"]["thought"]) > 0
            ):
                for thought_response in response_json["data"]["thought"]:
                    thoughts.append(Thought.from_json(thought_response))
            else:
                print("No thought sent for room with since filter")

        return thoughts

    def get_shared_room_id(self, user: User, mammut_user_id):

        room_found = None
        for user_room in user.rooms:
            # needs to get all info
            room = self.get_room(user_room.id)

            if room.created_by.id == mammut_user_id:
                return room
            else:
                for user in room.members:
                    if user.id == mammut_user_id:
                        return room

        return room_found

    def get_fixed_mammut(self):
        response = self.get_request(self.fixed_user_path)

        if response is None:
            print("Error getting fixed mammut user")

        else:
            response_json = response.json()

            # Get first user response
            user = User.from_json(response_json["data"]["user"][0])
            return user

    def get_room(self, room_id):
        room = None
        url = self.get_room_path.replace("{roomId}", str(room_id))
        response = self.get_request(url)

        if response is None:
            print("Error getting room: " + str(room_id))
        else:
            response_json = response.json()

            if (
                "room" in response_json["data"].keys()
                and len(response_json["data"]["room"]) > 0
            ):
                room = Room.from_json(response_json["data"]["room"][0])
            else:
                print("No room found for: " + str(room_id))

        return room

    def get_videos(self, mammut_id):
        videos = []
        url = self.get_videos_path.replace("{userId}", str(mammut_id))
        response = self.get_request(url)

        if response is None:
            print("Error getting videos form user: " + str(mammut_id))
        else:
            response_json = response.json()
            print(response_json)

            if (
                "video" in response_json["data"].keys()
                and len(response_json["data"]["video"]) > 0
            ):
                videos = response_json["data"]["video"]
            else:
                print("No videos found for: " + str(mammut_id))

        return videos

    def post_request(self, request_path: str, data: {}, data_is_json=False):
        url = request_path.replace("{mammutURL}", self.mammutUrl)
        if data_is_json:
            json_data = data
        else:
            json_data = json.dumps(data)
        response = requests.post(
            url, data=json_data, headers={"Content-type": "application/json"}
        )

        if response.status_code != 200:
            print("Error connecting to Mammut api. Code: " + str(response.status_code))

        return response

    def get_request(self, request_path: str):
        url = request_path.replace("{mammutURL}", self.mammutUrl)
        response = requests.get(url)

        if response.status_code != 200:
            print("Error connecting to Mammut api. Code: " + str(response.status_code))

        return response

    def post_compilation(self, payload):
        response = self.post_request(self.post_compilation_path, payload, True)
        if response is not None:
            if response.status_code == "200":
                return 200
            elif response.status_code == "503":
                return 503
            else:
                print("Unknown respose status code received")
                return int(response.status_code)
        else:
            print("Error posting compilation to API.")

    def get_unique_vertex(self, vertex: str, prop_name: str, prop_value: str):
        str_replacements_dict = {
            "{vertex}": vertex,
            "{prop_name}": prop_name,
            "{prop_value}": prop_value,
        }
        rest_url = self.get_vertex_by_unique_property_path.replace(
            "{mammutURL}", self.mammutUrl
        )
        for i, j in str_replacements_dict.items():
            rest_url = rest_url.replace(i, j)

        api_response = self.get_request(rest_url)

        if api_response is None:
            return {"error": {"message": "Error communicating with Mammut API"}}
        else:
            api_response_json = api_response.json()
            if (
                api_response_json["status"] == "Success"
                and len(api_response_json["data"][vertex]) > 0
            ):
                user = api_response_json["data"][vertex][0]
                return user
            else:
                return {}


class UsersLoader:
    DEFAULT_USERS_LOADER_START_CELL = "A1"
    DEFAULT_USERS_LOADER_START_COLUMN = "A"
    DEFAULT_USERS_LOADER_START_ROW = "1"
    DEFAULT_USERS_LOADER_END_CELL = "E"
    DEFAULT_MAIN_EMAIL_COLUMN_NAME = "main-email"
    DEFAULT_ID_COLUMN_NAME = "id"
    storage_manager = StorageManager()

    def __init__(self, spreadsheet_id: str, sheet_name: str, mammut_api: Request):
        self.spreadsheet_id = spreadsheet_id
        self.sheet_name, self.end_cell = sheet_name.split("-")
        self.mammut_api = mammut_api
        self.users_dataframe = None
        self.users_list = []

    def get_users_dataframe(self):

        self.users_dataframe = self.storage_manager.get_spreadsheet_as_dataframe(
            self.spreadsheet_id,
            self.sheet_name,
            UsersLoader.DEFAULT_USERS_LOADER_START_COLUMN,
            UsersLoader.DEFAULT_USERS_LOADER_START_ROW,
            self.end_cell,
        )

    def create_user(self, df_row):
        user = None
        payload = {col: df_row[col] for col in df_row.keys()}
        user_response = self.mammut_api.post_user(user, payload)
        if user_response is not None:
            payload["id"] = user_response
            return payload
        else:
            print("error creating user with payload")
            return None

    # TODO: al hacer el post de user deberia ahcer get del mismo para tener todos los campos??o con agregar los mismos del sheet base basta?
    def check_users(self):
        for index, row in self.users_dataframe.iterrows():
            raw_user = self.mammut_api.get_user_by_mail(
                row[UsersLoader.DEFAULT_MAIN_EMAIL_COLUMN_NAME], raw_user=True
            )
            if raw_user == None:
                raw_user = self.create_user(row)
                if raw_user is not None:
                    self.users_list.append(raw_user)
            else:
                self.users_list.append(raw_user)
        return self.users_list

    def save_users(self):
        sheet_titles = list()
        all_users_keys = list(map(lambda user: list(user.keys()), self.users_list))
        all_users_keys = reduce((lambda x, user_keys: x + user_keys), all_users_keys)
        unique_keys = list(set(all_users_keys))
        sheet_titles = sheet_titles + unique_keys

        user_keys = sheet_titles
        row = []
        rows = []
        rows.append(sheet_titles)
        for user in self.users_list:
            for key in user_keys:
                user_json = user
                if key in user_json:
                    if isinstance(user_json[key], dict):
                        row.append(json.dumps(user_json[key]))
                    else:
                        row.append(user_json[key])
                else:
                    row.append("empty")
            rows.append(row)
            row = []
        temporal_sheet_name = self.sheet_name + "-result"
        temporal_sheets = self.storage_manager.get_temporal_sheets_from_new_to_old(
            self.spreadsheet_id, temporal_sheet_name
        )
        if len(temporal_sheets) > 0:
            for ts in temporal_sheets:
                self.storage_manager.delete_sheet(self.spreadsheet_id, ts[0])
        dump_sheet_title = self.storage_manager.create_temporal_new_sheet(
            self.spreadsheet_id, temporal_sheet_name
        )
        dump_range_name = (
            dump_sheet_title + "!" + UsersLoader.DEFAULT_USERS_LOADER_START_CELL
        )
        self.storage_manager.add_rows_to_spreadsheet(
            self.spreadsheet_id,
            dump_sheet_title,
            rows,
            UsersLoader.DEFAULT_USERS_LOADER_START_ROW,
            UsersLoader.DEFAULT_USERS_LOADER_START_COLUMN,
        )

    def load_users(self):
        self.get_users_dataframe()
        self.check_users()
        self.save_users()


class RoomsCreator:
    DEFAULT_ROOM_CREATOR_START_CELL = "A1"
    DEFAULT_ROOM_CREATOR_START_COLUMN = "A"
    DEFAULT_ROOM_CREATOR_START_ROW = "1"
    DEFAULT_ROOM_CREATOR_END_CELL = "F"
    storage_manager = StorageManager()

    def __init__(self, mammut: User, spreadsheet_id, mammut_api):
        self.mammut = mammut
        self.spreadsheet_id = spreadsheet_id
        self.rooms_dataframe = None
        self.rooms = list()
        self.mammut_api = mammut_api

    def get_rooms(self, sheet_title):
        self.rooms_dataframe = self.storage_manager.get_spreadsheet_as_dataframe(
            self.spreadsheet_id,
            sheet_title,
            RoomsCreator.DEFAULT_ROOM_CREATOR_START_COLUMN,
            RoomsCreator.DEFAULT_ROOM_CREATOR_START_ROW,
            RoomsCreator.DEFAULT_ROOM_CREATOR_END_CELL,
        )

    def create_rooms(self, character_limit=None):
        date = mammut.get_str_format_timezone_aware_datetime_now()
        for index, row in self.rooms_dataframe.iterrows():
            title = row.title if not character_limit else row.title[:character_limit]
            room = Room(
                title, RoomType.group, date, self.mammut, row.url, row.custom_code
            )
            self.rooms.append(room)
            self.mammut_api.post_room(room)

    def save_rooms(self, sheet_title):
        sheet_titles = ["id", "title"]
        row = []
        rows = []
        rows.append(sheet_titles)
        for room in self.rooms:
            row.append(room.id)
            row.append(room.title)
            row.append(room.url)
            row.append(room.custom_code)
            rows.append(row)
            row = []
        temporal_sheet_name = sheet_title + "-rooms"
        temporal_sheets = self.storage_manager.get_temporal_sheets_from_new_to_old(
            self.spreadsheet_id, temporal_sheet_name
        )
        if len(temporal_sheets) > 0:
            for ts in temporal_sheets:
                self.storage_manager.delete_sheet(self.spreadsheet_id, ts[0])
        dump_sheet_title = self.storage_manager.create_temporal_new_sheet(
            self.spreadsheet_id, temporal_sheet_name
        )
        dump_range_name = (
            dump_sheet_title + "!" + UsersLoader.DEFAULT_USERS_LOADER_START_CELL
        )
        self.storage_manager.add_rows_to_spreadsheet(
            self.spreadsheet_id,
            dump_sheet_title,
            rows,
            UsersLoader.DEFAULT_USERS_LOADER_START_ROW,
            UsersLoader.DEFAULT_USERS_LOADER_START_COLUMN,
        )

    def activate(self, sheet_title):
        self.get_rooms(sheet_title)
        self.create_rooms()
        self.save_rooms(sheet_title)



from mammut.common.simulator_utilities import *
from ipywidgets import *
from django.template.loader import get_template

from mammut.common.mammut_api import UserType, RoomType, Thought, Issue


def get_simulation_boxes(simulator_obj, move_boxes_callback):
    # Select your mammut layout #
    mammut_user_dropdown = widgets.Dropdown(
        description="Mammut bot", disabled=False, button_style=""
    )
    mammut_user_dropdown.options = simulator_obj.users[UserType.machine.value]
    select_mammut_user_btn = widgets.Button(
        description="Seleccionar", button_style="primary"
    )

    def select_user_button_clicked(b):
        simulator_obj.current_user = simulator_obj.mammut_api.get_user(
            mammut_user_dropdown.value
        )
        move_boxes_callback()

    select_mammut_user_btn.on_click(select_user_button_clicked)
    select_mammut_box = widgets.VBox(
        [widgets.HBox([mammut_user_dropdown, select_mammut_user_btn])]
    )
    # /Select your mammut layout #

    # Select person user #
    select_person_user_dropdown = widgets.Dropdown(
        description="Su usuario:", disabled=False, button_style=""
    )
    select_person_dict = simulator_obj.users[UserType.person.value]
    select_person_dict.update(simulator_obj.users[UserType.normal.value])
    select_person_user_dropdown.options = select_person_dict
    select_user_btn = widgets.Button(description="Seleccionar", button_style="primary")

    def select_person_user_clicked(b):
        simulator_obj.current_person_user = simulator_obj.mammut_api.get_user(
            select_person_user_dropdown.value
        )
        move_boxes_callback()

    select_user_btn.on_click(select_person_user_clicked)
    select_user_box = widgets.VBox(
        [widgets.HBox([select_person_user_dropdown, select_user_btn])]
    )
    # /Select person user #

    # Select conversation room #
    select_room_dropdown = widgets.Dropdown(
        description="Conversation room:", disabled=False, button_style=""
    )
    rooms_dict = simulator_obj.rooms[RoomType.private.value]
    rooms_dict.update(simulator_obj.rooms[RoomType.group.value])
    select_room_dropdown.options = rooms_dict
    select_room_btn = widgets.Button(description="Seleccionar", button_style="primary")

    def select_room_clicked(b):
        simulator_obj.chat_room = simulator_obj.mammut_api.get_room(
            select_room_dropdown.value
        )
        move_boxes_callback()

    select_room_btn.on_click(select_room_clicked)
    select_room_box = widgets.VBox(
        [widgets.HBox([select_room_dropdown, select_room_btn])]
    )
    # /Select conversation room #

    # Send conversation messages #
    thoughts_list_template = get_template("thoughts_list.html")
    context_dict = {"thoughts_list": ["Hey bot", "Hello you!"]}
    thoughts_list_html = thoughts_list_template.render(context_dict)
    thoughts_list_widget = widgets.HTML(value=thoughts_list_html)
    send_message_textbox = widgets.Text(
        value="Hola Mammut", description="Mensaje:", layout=Layout(width="100%")
    )
    send_message_btn = widgets.Button(
        description="Enviar mensaje", button_style="primary"
    )
    get_thoughts_btn = widgets.Button(
        description="refresh thoughts", button_style="primary"
    )
    issue_is_present_output_alert = widgets.HTML(value="")

    def get_thouhts_btn_clicked(b):
        thoughts = simulator_obj.mammut_api.get_thoughts(
            simulator_obj.chat_room.id, from_type="Room"
        )
        room_thoughts = thoughts
        print("mensajes", thoughts)
        thoughts_text = [thought.text for thought in thoughts]
        print(thoughts)
        context_dict = {"thoughts_list": thoughts_text}
        thoughts_list_html = thoughts_list_template.render(context_dict)
        thoughts_list_widget.value = thoughts_list_html

        if room_thoughts[len(room_thoughts) - 1].issue is not None:
            issue_is_present_output_alert.value = (
                "<h4>There is a new Issue in the conversation</h4>"
            )
        elif (
            len(room_thoughts) > 1
            and room_thoughts[len(room_thoughts) - 2].issue is not None
        ):
            issue_is_present_output_alert.value = (
                "<h4>There is a new Issue in the conversation</h4>"
            )
        else:
            issue_is_present_output_alert.value = ""

    get_thoughts_btn.on_click(get_thouhts_btn_clicked)

    def send_message_btn_clicked(b):
        message = send_message_textbox.value
        test_thought = Thought(
            simulator_obj.current_person_user.id,
            simulator_obj.chat_room.id,
            message,
            mammut.get_str_format_timezone_aware_datetime_now(),
        )
        simulator_obj.mammut_api.post_thought(test_thought)

    send_message_btn.on_click(send_message_btn_clicked)
    send_message_box = widgets.VBox(
        [
            thoughts_list_widget,
            issue_is_present_output_alert,
            get_thoughts_btn,
            send_message_textbox,
            send_message_btn,
        ],
        layout=Layout(width="100%", align_items="center"),
    )

    # /Send conversation messages #

    # Solve conversation issues #
    ignore_issue_html = widgets.HTML("<h2>Ignorar Issue</h2>")
    ignore_issue_label = widgets.Label(
        value="Ignorar y desbloquear conversacion",
        layout=Layout(width="100%", align_items="center"),
    )
    ignore_issue_button = widgets.Button(description="Ignorar", button_style="primary")
    ignore_issue_output_html = widgets.HTML("")

    def ignore_issue_button_clicked(b):
        issue = Issue(
            room_id=simulator_obj.chat_room.id,
            issue_type="RemoveHaltedMark",
            machine_id=simulator_obj.current_user.id,
            causative_connected=False,
            resolution_text="",
            issue_state="open",
            custom_code="SomeCustomCode",
        )
        res = simulator_obj.mammut_api.post_remove_halted_mark_issue(issue)
        if res.status_code == 200:
            ignore_issue_output_html.value = "<p>Issue ignorado correctamente!</p>"
        else:
            ignore_issue_output_html.value = (
                "<p>Hubo un problema al intentar ignorar el Issue!</p>"
            )

    ignore_issue_button.on_click(ignore_issue_button_clicked)
    ignore_issue_box = widgets.VBox(
        [
            ignore_issue_html,
            ignore_issue_label,
            ignore_issue_button,
            ignore_issue_output_html,
        ],
        layout=Layout(width="100%", align_items="center"),
    )

    solve_issues_box = widgets.VBox(
        [widgets.HBox([ignore_issue_box])], layout=Layout(width="100%")
    )

    # /Solve conversation issues #

    return tuple(
        [
            select_mammut_box,
            select_user_box,
            select_room_box,
            send_message_box,
            solve_issues_box,
        ]
    )
